#version 120

varying vec3 frag_position;					// in object space
varying vec4 pos;							// from vertex shader

uniform sampler3D volume_texunit;			// volume texture
uniform sampler1D trfunc_texunit;			// transfer function

uniform sampler2D particle_depth_texunit;	// particle depth texture for visibility check 
uniform sampler2D particle_color_texunit;	// particle color texture
uniform sampler3D gradient_texunit;			// texture with the gradient for shading computation

uniform vec3 light_position[4];				// one light for shading computation
uniform bool light_enable[4];

uniform vec3 eye_position;					// camera position in object space

uniform float sample_step;					// step used to advance the sampling ray

uniform float shading_intensity;			// intensity of shading in percent
uniform float gradient_intensity;			
uniform float opacity_intensity;			

uniform bool particles_enabled;

uniform vec3 gradMin;
uniform vec3 gradMax;

// function header
vec4 computeFragColor(vec3 iso_pos, vec4 color);
vec3 blinn(vec3 N, vec3 V, vec3 L, int light);

void main(void)
{
	// ray direction goes from frag_position in the _inverse_ eye_position, i.e. front to back
	vec3 ray_dir = normalize(frag_position - eye_position);
	vec3 ray_pos = gl_TexCoord[0].xyz;

	// entry and exit position
	vec3 pos111 = vec3(1.0, 1.0, 1.0);
	vec3 pos000 = vec3(0.0, 0.0, 0.0);

	// get particle color from the texture
	vec2 texC = pos.xy/pos.w;
	texC = 0.5 * texC + vec2(0.5);

	vec4 particleColor = texture2D(particle_color_texunit, texC);

	// REMEMBER TO CHANGE THIS IF WE CHANGING THE PROJECTION MATRIX!!!!
	float n = 1.0;		// near
	float f = 100.0;	// far

	// get the depth of the pixel from the camera see: http://olivers.posterous.com/linear-depth-in-glsl-for-real
	float depth = texture2D(particle_depth_texunit, texC).r;
	float z_n = 2.0 * depth - 1.0;
	depth = 2.0 * (n * f) / (f + n - z_n * (f - n));

	vec4 frag_color = vec4(0.0, 0.0, 0.0, 0.0);
	vec4 color;

	do
	{
		ray_pos += ray_dir * sample_step;

		// break out if ray reached the end of the cube
		if(any(greaterThan(ray_pos, pos111)))
		{
			break;
		}

		// break out if ray reached the end of the cube
		if(any(lessThan(ray_pos, pos000)))
		{
			break;
		}

		// early ray termination
		if (frag_color.a >= 1.0)
		{
			break;
		}

		// check for particles
		vec4 tmp = gl_ProjectionMatrix * gl_ModelViewMatrix * vec4((ray_pos - 0.5) * 20.0, 1.0);

		if(depth != 1.0 && depth < tmp.w && particles_enabled)
		{
			// just compose the particle color -> no further volume shading here. Particles should be shaded in the other fragment shader
			frag_color.rgb = frag_color.rgb + (1.0 - frag_color.a) * particleColor.rgb;

			gl_FragColor = clamp(vec4(frag_color.rgb, 1.0), vec4(0), vec4(1));

			return;
		}

		float density = texture3D(volume_texunit, ray_pos).r;

		color = texture1D(trfunc_texunit, density);

		if (shading_intensity > 0) {
			color = computeFragColor(ray_pos, color);	// we need to accumulate the color each step!
		}

		// front to back compositing
		frag_color.rgb = frag_color.rgb + (1.0 - frag_color.a) * color.rgb * color.a;
		frag_color.a = frag_color.a + (1.0 - frag_color.a) * color.a;
	}
	while(true);

	// check for particles which are behind the volume (i.e. on the boundary)
	if(depth != 1.0 && particles_enabled) 
	{
		frag_color.rgb = frag_color.rgb + (1.0 - frag_color.a) * particleColor.rgb;

		gl_FragColor = clamp(frag_color, vec4(0), vec4(1));

		return;
	}

	if(frag_color == vec4(0.0, 0.0, 0.0, 0.0))
	{
		discard;
	}
	
	gl_FragColor = clamp(frag_color, vec4(0), vec4(1));
}

vec3 blinn(vec3 N, vec3 V, vec3 L, int light)
{
	//Not used
	//vec3 Ka = vec3(1.0, 1.0, 1.0);
	//vec3 Kd = vec3(1.0, 1.0, 1.0);
	//vec3 Ks = vec3(1.0, 1.0, 1.0);

	float shininess = 100.0;

	float diff_coeff = max(dot(L, N), 0.0);
	
	vec3 H = normalize(L + V);

	if(gl_LightSource[light].specular.r == 0 &&  gl_LightSource[light].specular.g == 0.0 && gl_LightSource[light].specular.b == 0.0)
	{
		return (gl_LightSource[light].ambient.rgb + gl_LightSource[light].diffuse.rgb * diff_coeff);
	}
	else	{
		float spec_coeff = pow(max(dot(H, N), 0.0), shininess);

		if (diff_coeff <= 0.0) 
		{
			spec_coeff = 0.0;
		}

		spec_coeff = 0.0;

		return (gl_LightSource[light].ambient.rgb + gl_LightSource[light].diffuse.rgb  * diff_coeff + gl_LightSource[light].specular.rgb * spec_coeff);
	}

	//return (Ka * gl_LightSource[light].ambient.rgb + Kd * gl_LightSource[light].diffuse.rgb  * diff_coeff + Ks * gl_LightSource[light].specular.rgb * spec_coeff);
	//return (Ka + Kd * diff_coeff + Ks * spec_coeff);
}

vec4 computeFragColor(vec3 iso_pos, vec4 color)
{
	// get the 'real' gradient
	vec3 grad = texture3D(gradient_texunit, iso_pos).xyz * (gradMax - gradMin) + gradMin;

	float gradient_magnitude = length(grad);

	vec3 N = normalize(grad);

	vec3 V = normalize(eye_position - frag_position);

	vec3 shaded_color = color.rgb;

	float shading_intensity_calc = shading_intensity * 20.0;

	//Only one light source
	for(int i=0; i<1; i++)
	{
		//if (light_enable[i])
		{
			vec3 L = normalize(light_position[i] - frag_position);

			if (dot(L, N) < 0.0)
			{
				L = -L;
			}

			shaded_color = shaded_color + shaded_color * blinn(N, V, L, i) * shading_intensity_calc;
		}
	}

	vec4 return_color = vec4(shaded_color, color.a);

	//Gradient-Magnitude Shading Modulation
	float t = smoothstep(0.0, 1.0 - gradient_intensity, gradient_magnitude);

	return_color = mix(color, return_color, t);

	//Gradient-Magnitude Opacity Modulation
	t = smoothstep(0.0, 1.0 - opacity_intensity, gradient_magnitude);

	return_color.a *= t;

	return return_color;
}